import logging

import convo_sdk.version

logger = logging.getLogger(__name__)

__version__ = convo_sdk.version.__version__

import convo_sdk.cli
from convo_sdk.interfaces import Tracker, Action, ActionExecutionRejection

if __name__ == "__main__":
    import convo_sdk.__main__

    convo_sdk.__main__.main()
